@@ -1,38 +1,47 @@
 import numpy as np
-from numba import njit, types
-from .core import find_indices
+from numba import njit, types, prange, typed
+import pickle
 import os
+from numpy import ndarray
 
 
-def load_sparse(data_path: str, coords_path='', shape='', mode='r'):
+def load_sparse(data_path: str, coords_path=None, coords_dict_path = None, shape=None):
     """
     Load a (memory-mapped) sparse array from disk
     :param data_path: path to sparse data array OR parent directory containing 'sparse_data.npy', 'sparse_coords.npy', and 'dense_shape.npy' arrays
-    :param coords_path: path to sparse coordinates array
-    :param shape: shape of the dense array, as either tuple or path to numpy array containing shape
-    :param mode: mode to open the sparse arrays in (e.g., read-only, read-write, etc.) - r/r+/w
+    :param coords_path: (optional) path to sparse coordinates array
+    :param coords_dict_path: (optional) path to dictionary mapping sparse to dense row coordinates
+    :param shape: (optional) shape of the dense array, as either tuple or path to numpy array containing shape
     :return: SparseArray object
     """
     if os.path.isdir(data_path):
         coords_path = os.path.join(data_path, 'sparse_coords.npy')
+        coords_dict_path = os.path.join(data_path, 'sparse_coords_dict.pkl')
         shape = os.path.join(data_path, 'dense_shape.npy')
         data_path = os.path.join(data_path, 'sparse_data.npy')
 
-    return SparseArray(data_path, coords_path, shape, mode)
+    return SparseArray(data_path, coords_path, coords_dict_path, shape, mode='r')
 
 
 class SparseArray:
-    def __init__(self, data_path: str, coords_path: str, shape: str or tuple, mode='r'):
+    def __init__(self, data_path: str, coords_path: str, coords_dict_path, shape: str or tuple, mode='r'):
         self.data = np.load(data_path, mmap_mode=mode)
         self.data_dtype = self.data.dtype
 
         self.coords = np.load(coords_path, mmap_mode=mode)
         self.coords_shape = self.coords.shape
+        with open(coords_dict_path, 'rb') as f:
+            self.coords_dict = typed.Dict.empty(key_type=types.int64, value_type=types.UniTuple(types.int64, count=2))
+            coords_dict = pickle.load(f)
+            for key, value in coords_dict.items():
+                self.coords_dict[key] = value
 
         if shape is tuple:
             self.dense_shape = shape
+            self.shape = self.dense_shape
         else:
             self.dense_shape = tuple(np.load(shape))
+            self.shape = self.dense_shape
 
     def __getitem__(self, items):
         """"
@@ -117,8 +126,11 @@ class SparseArray:
                            np.ndarray: self.__find_rows_ndarray}
 
         # Find the target coordinates for the given row indices
-        find_coords, coords_present = search_function[type(row_idx)](self.coords, row_idx,
-                                                                     maxlen=self.coords_shape[0])
+        find_coords, coords_present = search_function[type(row_idx)](self.coords_dict, row_idx)
+
+        find_coords = np.concatenate(find_coords)
+        find_coords = find_coords[find_coords >= 0]
+        find_coords = np.sort(find_coords)
 
         # Assuming the row indices have non-zero data...
         if coords_present:
@@ -127,67 +139,70 @@ class SparseArray:
 
             # Some slightly complex indexing to identify the non-zero cells of the (indexed) target array
             nonzero_row_indices_unique = np.where(np.isin(row_idx, np.unique(target_coords[:, 0])))[0]
-            nonzero_row_indices_unique_idx = np.unique(target_coords[:, 0], return_index=True)[1]
+            nonzero_row_indices_unique_idx = np.where(np.isin(np.unique(target_coords[:, 0]), row_idx))[0]
 
-            target_coords[:, 0] = -1
+            target_coords[:, 0] = 0
             target_coords[nonzero_row_indices_unique_idx, 0] = nonzero_row_indices_unique
 
             target_coords[:, 0] = np.maximum.accumulate(target_coords[:, 0], axis=0)
 
             # Re-create the dense array (of the shape requested) with all zeros
-            target_data = np.zeros((1,) + self.dense_shape[1:]) if isinstance(row_idx, int) else np.zeros(
-                (len(row_idx),) + self.dense_shape[1:])
+            target_data = np.zeros((1,) + self.dense_shape[1:], dtype=self.data_dtype) if isinstance(row_idx, int) else np.zeros(
+                (len(row_idx),) + self.dense_shape[1:], dtype=self.data_dtype)
 
             # Fill the dense array with the non-zero data at the target coordinates
             target_data[tuple(target_coords.T)] = self.data[find_coords]
         else:
             # If the row indices have no non-zero data, return an array of zeros
-            target_data = np.zeros((1,) + self.dense_shape[1:]) if isinstance(row_idx, int) else np.zeros(
-                (len(row_idx),) + self.dense_shape[1:])
+            target_data = np.zeros((1,) + self.dense_shape[1:], dtype=self.data_dtype) if isinstance(row_idx, int) else np.zeros(
+                (len(row_idx),) + self.dense_shape[1:], dtype=self.data_dtype)
 
         return target_data
 
     @staticmethod
-    @njit(types.Tuple((types.int64[:], types.boolean))(types.Array(types.int32, 2, 'C', readonly=True), types.int64,
-                                                       types.int64))
-    def __find_rows_int(coords: np.ndarray, row_idx: int, maxlen: int) -> types.Tuple:
-        """
-        Efficiently search the coordinates to find the indices that match the given row index
-        :param coords: numpy 2D array of coordinates
-        :param row_idx: target row index (integer)
-        :param maxlen: maximum length of the coordinates array
-        :return: numpy array of coordinates that match the given row index
-        """
-
-        # Start with an efficient binary tree search algorithm for identifying the start and end indices
-        start_point, end_point = find_indices(coords, row_idx, 0, maxlen - 1)
-
-        coords_present = True if start_point >= 0 else False
-
-        return np.array([i for i in range(start_point, end_point + 1)]), coords_present
-
-    @staticmethod
-    @njit(types.Tuple((types.int64[:], types.boolean))(types.Array(types.int32, 2, 'C', readonly=True),
-                                                       types.Array(types.int64, 1, 'A'), types.int64))
-    def __find_rows_ndarray(coords: np.ndarray, row_idx: np.ndarray, maxlen: int) -> (np.ndarray, bool):
+    @njit(types.Tuple((types.List(types.Array(types.int64, 1, 'C')),
+                       types.boolean))(types.DictType(types.int64, types.UniTuple(types.int64, count=2)),
+                                       types.int64))
+    def __find_rows_int(coords_dict, row_idx: int) -> tuple[list[ndarray], bool]:
         """
         Efficiently search the coordinates to find the indices that match the given row indices
-        :param coords: numpy 2D array of coordinates
-        :param row_idx: target row indices (numpy array)
-        :param maxlen: maximum length of the coordinates array
+        :param coords_dict: dictionary mapping each dense row index to the first/last indices of the coordinates
+        :param row_idx: target row index (integer)
         :return: numpy array of coordinates that match the given row indices
         """
 
-        find_coords = []
+        start, end = coords_dict[row_idx]
+        find_coords = [np.arange(start, end+1)] if start >= 0 else [np.array([-1])]
+        coords_present = True if start >= 0 else False
 
-        for row in np.sort(row_idx):
-            # Start with an efficient binary tree search algorithm for identifying the start and end indices
-            start_point, end_point = find_indices(coords, row, 0, maxlen - 1)
-            find_coords.extend([i for i in range(start_point, end_point + 1)]) if start_point >= 0 else None
+        return find_coords, coords_present
 
-        coords_present = True if find_coords else False
+    """
+        @njit(types.Tuple((types.List(types.Array(types.int64, 1, 'C')), types.boolean))(types.DictType(types.int64, types.UniTuple(types.int64, count=2)),
+                                                                                         types.Array(types.int64, 1, 'A')),
+              parallel=True)
+    """
+    @staticmethod
+    def __find_rows_ndarray(coords_dict, row_idx: np.ndarray) -> tuple[list[ndarray], bool]:
+        """
+        Efficiently search the coordinates to find the indices that match the given row indices
+        :param coords_dict: dictionary mapping each dense row index to the first/last indices of the coordinates
+        :param row_idx: target row indices (numpy array)
+        :return: numpy array of coordinates that match the given row indices
+        """
 
-        return np.array(find_coords), coords_present
+        find_coords = [np.array([-1]) for _ in range(len(row_idx))]
+        coords_present = False
+
+        for row in prange(len(row_idx)):
+            start, end = coords_dict[row_idx[row]]
+            find_coords[row] = np.arange(start, end+1) if start >= 0 else np.array([-1])
+
+            coords_present += True if start >= 0 else False
+
+        coords_present = bool(coords_present)
+
+        return find_coords, coords_present
 
     def __get_item(self, items) -> np.ndarray:
         """
@@ -206,3 +221,6 @@ class SparseArray:
 
         return target_data
 
+    def __len__(self):
+        return self.shape[0]
+
